#  -*- coding: utf-8 -*-
#
#  GSXML - A Content Import/Export Product
#
#  Uses a specific context to import/export marshalled data
#
#  Copyright (c) 2006-2007 InQuant GmbH
#
#  Parts of this code is ruthlessy stolen from XNSMarshaller,
#  a fork of the Marshaller Product in collective. Original
#  work done by:
#
#  kapil thangavelu <k_vertigo@objectrealms.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
""" the content uses a context for data storage and is responsible
for marshalling
"""

__author__ = """Ramon Bartl <ramonski@ramonski.de>"""
__docformat__ = 'restructuredtext'
__revision__ = "$Revision: 43706 $"

import os
import logging
from StringIO import StringIO
import pickle

import exceptions

from zope.interface import implements

from OFS.Image import File
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import _createObjectByType
from Products.Archetypes.interfaces import IBaseFolder
from Products.Archetypes.public import FileField
from Products.Archetypes.utils import getRelURL
from Products.Marshall.registry import getComponent
from interfaces import IXMLMarshaller, IXMLDemarshaller, IGSXMLFilesystemExporter, IGSXMLFilesystemImporter
from inquant.plone.gsxml import manifest, collector, config

import utils

LOGGER="gsxml.content"

def info(msg):
    logging.getLogger(LOGGER).info(msg)

def debug(msg):
    logging.getLogger(LOGGER).debug(msg)

def error(msg):
    logging.getLogger(LOGGER).error(msg)

MANIFEST = "MANIFEST.XML"

#
#   setup_tool handlers
#
def exportSiteStructure(context):
    IGSXMLFilesystemExporter(context.getSite()).export(context, 'structure', True)

def importSiteStructure(context):
    IGSXMLFilesystemImporter(context.getSite()).import_(context, 'structure', True)

class GSXMLImportError(exceptions.RuntimeError):
    pass

class ATCTXMLMarshaller(object):
    """
    Marshalls ATContentType -> XML
    """
    implements( IXMLMarshaller )

    def __init__(self, context ):
        info( "ATCTXMLMarshaller.__init__: context=%s" % (context.__class__) )
        self.context = context
        self.marshaller = getComponent( "atxml" )
        self.options = {}

    def getxml(self):
        info( "ATCTXMLMarshaller.getxml" )
        content_type, length, data = self.marshaller.marshall( self.context, **self.options )
        self._xml = data
        self._content_type = content_type
        self._length = length
        return self._xml

    xml = property(getxml, "Property of 'xml'")

class ATCTXMLDemarshaller(object):
    """
    Adapter which adapts any ATContentType to IXMLDemarshaller
    """
    implements( IXMLDemarshaller )

    def __init__( self, context ):
        info( "ATCTXMLDemarshaller.__init__: context=%s" % ( context ) )
        self.context = context
        self.marshaller = getComponent( "atxml" )
        self.options = {}

    def demarshall( self, instance, xml ):
        info( "ATCTXMLDemarshaller.demarshall: instance=%s" % ( instance ) )
        self.marshaller.demarshall( instance, xml )
        instance.reindexObject()

class XMLContentRootExporter(object):
    implements( IGSXMLFilesystemExporter )

    def __init__( self, context ):
        self.context = context

        self.collector = collector.SimpleDictCollector()
        self.fse = XMLContentFSExporter( context, self.collector ) 
        self.manifest = manifest.ManifestWriter( StringIO() )

    def write_manifest( self, export_context, subdir ):
        stream = StringIO()
        self.collector.getXml( stream )
        export_context.writeDataFile( MANIFEST, text=stream.getvalue(),
                content_type="text/xml", subdir=subdir)
        export_context.writeDataFile( "MANIFEST.pickle", text=self.collector.getData(),
                content_type="application/octet-stream", subdir=subdir)

    def write_dot_objects( self, export_context, subdir ):
        self.manifest.add( self.context.getId(), self.context.getPortalTypeName()  )
        export_context.writeDataFile('.objects.xml',
                                     text=self.manifest.getvalue(),
                                     content_type='text/xml',
                                     subdir=subdir
                                     )

    # IFilesystemExporter
    def export( self, export_context, subdir, root=False ):
        info( "XMLContentRootExporter.export: export_context=%s, subdir=%s, root=%s" % (
                export_context, subdir, root ) )

        # top-level .objects.xml
        self.write_dot_objects( export_context, subdir )

        # start export
        self.fse.export( export_context, subdir, root=False )

        # write export manifest
        self.write_manifest( export_context, subdir )

    def listExportableItems(self):
        return ()

class XMLContentFSExporter(object):
    implements( IGSXMLFilesystemExporter )

    def __init__( self, context, collector=None ):
        info( "XMLContentFSExporter.__init__: context=%s" % ( context ) )
        self.context = context
        self.ref_catalog = getToolByName(self.context, 'reference_catalog')
        self.collector = collector

        # True if folder
        self.folderish = IBaseFolder.providedBy( context )
        info( "XMLContentFSExporter: context folderish: %s" % self.folderish )

        self.manifest = manifest.ManifestWriter( StringIO() )


    # IGSXMLFilesystemExporter
    def export( self, export_context, subdir, root=False ):
        info( "XMLContentFSExporter.export: export_context=%s, subdir=%s, root=%s" % (
                export_context, subdir, root ) )

        # fetch an marshaller for this context
        adapter = IXMLMarshaller( self.context )

        if self.collector is not None:
            self.collector.addObjectInfo( self.context.UID(),
                    subdir=subdir,
                    UID=self.context.UID(),
                    id=self.context.getId(),
                    type=self.context.portal_type )

        # marshall binary data and record the field names for which we did so
        atns_exclude = self.exportFiles( export_context, subdir )

        # tell the marshaller to ignore the fields we exported the binary data
        adapter.options["atns_exclude"] = self.exportFiles( export_context, subdir )

        if config.NO_UID_ON_EXPORT:
            info( "GSXML: NO_UID_ON_EXPORT is True: UIDs wil NOT be exported." )
            adapter.options["atns_exclude"].append( "uid" )

        # marshall references
        ref_info = self.exportReferences( export_context, subdir )
        if self.collector is not None:
            self.collector.addRefInfo( self.context.UID(), ref_info )

        # Marshall our context
        fname = "%s.xml" % self.context.id
        path = os.path.join( subdir, fname )
        export_context.writeDataFile( path, adapter.xml, "text/xml" )

        if not self.folderish:
            return

        if not root:
            subdir = '%s/%s' % (subdir, self.context.getId())
            info( "XMLContentFSExporter: root! subdir=%s" % subdir )

        # Because we are a folder, export all our contents, too.
        # for this we need to fetch a list of objects which are
        # containd in the context, and export those for which
        # we can get and adapter to IGSXMLFilesystemExporter
        exportable = self.listExportableItems()

        for id, object, fse in exportable:
            if fse is None:
                continue
            # XXX: yuck!
            fse.collector = self.collector
            fse.export( export_context, subdir, False )
            self.manifest.add( id, object.getPortalTypeName()  )

        self.write_dot_objects( export_context, exportable, subdir )

    def exportReferences( self, export_context, subdir):
        """
        Export the references stored in our context.
        """
        mime ="application/octet-stream"
        
        # pickle the references
        data = pickle.dumps( self.context.at_references.aq_base )

        # contrieve a sensible filename for the reference pickle
        filename = "%s.refs.pickle" % self.context.getId()

        # write the stuff out to the storage backend
        export_context.writeDataFile(
                filename,
                data,
                mime,
                subdir
                   )
        if self.collector is not None:
                self.collector.addFileInfo( self.context.UID(), filename, mime )

        # now get some information about the references and
        # return them
        out = {}
        for ref in self.context.at_references.objectValues():
            ruid, info = self._getRefobjectInfo( ref )
            out[ruid] = info
        return out

    def _getRefobjectInfo( self, ref ):
        """
        Return a dict of information about one reference object.
        """
        return ref.UID(), dict(
                UID=ref.UID(),
                targetUID=ref.targetUID,
                sourceUID=ref.sourceUID,
                relationship=ref.relationship )


    def write_dot_objects( self, export_context, exportable, subdir ):
        # export xml manifesttns_exclude.append( field.getName() )
        export_context.writeDataFile('.objects.xml',
                                     text=self.manifest.getvalue(),
                                     content_type='text/xml',
                                     subdir=subdir
                                     )

    def exportFiles( self, export_context, subdir):
        """
        export any file based fields that aren't of type text/plain
        and return a list of field names exported in this fashion.
        """
        atns_exclude = []

        for field in self.context.Schema().fields():
            if isinstance( field, FileField ):
                mime = field.getContentType( self.context )
                if mime.startswith('text'):
                    continue

                atns_exclude.append( field.getName() )

                value = field.getRaw( self.context )
                if isinstance( value, File):
                    value = str(value.data)

                filename = "%s.%s.data"%( self.context.getId(),
                                 field.getName() )

                export_context.writeDataFile(
                        filename,
                        value,
                        mime,
                        subdir
                    )

                if self.collector is not None:
                    self.collector.addFileInfo( self.context.UID(), filename, mime )

        return atns_exclude

    def listExportableItems(self):
        """
        Return a list of tuples of exportable items:
          ( id, object, IGSXMLFilesystemExporter(obj) )

        i.e.
          [ ( "anobjectid", <ATDocument object ...>, <ATDocumentFilesystemExporter object ...> ), ( "nonexportable", <object>, None ), ... ]
        """
        info( "listExportableItems: context.id=%s" % ( self.context.id ) )
        if self.folderish:
            exportable = self.context.objectItems()
            exportable = [x + (IGSXMLFilesystemExporter(x[1], None),) for x in exportable]
            info( "listExportableItems: exportable=%(exportable)s" % locals() )
            return exportable
        else:
            return ()


class XMLContentRootImporter(object):
    """
    A controller for the import process.

    - reads and parses the Manifest
    - imports stuff
    - fix up imported objects
    """
    implements( IGSXMLFilesystemImporter )

    def __init__(self, context ):
        # TODO: ensure context is folderish.
        self.context = context

    def _create_collector( self, import_context, subdir ):
        self.manifest = import_context.readDataFile( MANIFEST, subdir )
        if manifest is None:
            return

        # XXX: too lazy to parse xml for now.
        data = import_context.readDataFile( "MANIFEST.pickle", subdir )
        if data is None:
            return

        self.collector = collector.SimpleDictCollector()
        self.collector.restoreData( data )

    def _fixup_import( self, import_context ):
        """
        Fixup the import.

        Here we ought to check whether or not reference UIDs changed during
        import. If so, we need to fix them. The sourceUID of the references
        is already fixed, the import step did that. So, we only need to
        check if reference target UIDs still match.

        We need to check every reference in the collector's reference map (which
        got built by the collector on import time) and check if a targetUID of those
        references has changed.
        """

        if self.collector is None:
            return # no fixup w/o collector info

        refmap = self.collector.getRefMap()

        ref_source_uids = refmap.keys()

        for sid, rm in refmap.iteritems():
            # XXX: this is lame. We need an adapter which adapts the changed reference to
            #      an conflict solver of some sort.
            if rm["new"] != rm["old"]:
                info( "GSXML IMPORT: reference %(ref)s (%(url)s): sourceUID %(old)s -> %(new)s" % rm )

                target = rm["target"]

                if target in ref_source_uids:
                    old = refmap[target]["old"]
                    new = refmap[target]["new"]
                    info( "GSXML IMPORT: FIXUP REFS: reference %(ref)s %(target)s -> %(new)s" % dict( ref=rm["ref"], new=new, target=target ) )

    def objects( self, import_context, subdir ):
        return manifest.ManifestReader( import_context.readDataFile('.objects.xml', subdir ) )

    def import_object( self, import_context, subdir, object_id, type_name, root=False ):
        """
        Import an object. Data is read from import_context using "subdir" and "object_id".
        """
        existing = self.context.objectIds()
        info( "XMLContentRootImporter.import_object: existing=%r" % ( list( existing ) ) ) 

        info( "XMLContentRootImporter.import_object: object_id=%s, type_name=%s" % ( object_id, type_name ) )

        if object_id not in existing:
            id = _createObjectByType(type_name, self.context, object_id)
            if id is None:
                error( "XMLContentRootImporter: Could not create instance of type %s, id %s." % ( type_name, object_id ) )
                raise GSXMLImportError( "Could not create instance of type %s, id %s." % ( type_name, object_id ) )
            info( "XMLContentRootImporter: Created object type %s, id %s." % ( type_name, object_id ) )

        object = self.context._getOb(object_id)

        adapter = IGSXMLFilesystemImporter(object)
        adapter.collector = self.collector

        adapter.import_(import_context, subdir, root=False)

    def import_( self, import_context, subdir, root=False ):
        # create a collector object
        self._create_collector( import_context, subdir )

        for object_id, type_name in self.objects( import_context, subdir ):
            self.import_object( import_context, subdir, object_id, type_name, root=False ) 

        # fixup import
        self._fixup_import( import_context )
        

class XMLContentFSImporter(object):
    implements(IGSXMLFilesystemImporter)

    def __init__( self, context, collector=None ):
        info( "XMLContentFSImporter.__init__: context=%s" % ( context ) )
        # this is the object which was newly created or is to be updated
        self.context = context

        # now lets look up adapter which provides IXMLDemarshaller for
        # this object:
        self.demarshaller = IXMLDemarshaller( self.context )

        # store the collector
        self.collector = collector

        # reference catalog for ref munging
        self.ref_catalog = getToolByName(self.context, 'reference_catalog')

        self.folderish = IBaseFolder.providedBy( context )
        info( "XMLContentFSImporter: context folderish: %s" % self.folderish )

    def objects( self, import_context, subdir ):
        return manifest.ManifestReader( import_context.readDataFile('.objects.xml', subdir ) )

    # IGSXMLFilesystemImporter
    def import_( self, import_context, subdir, root=False ):
        info( "XMLContentFSImporter.import: import_context=%s, subdir=%s, root=%s" % ( import_context, subdir, root ) )

        id = self.context.getId()
        filename = "%s.xml" % id
        xml = import_context.readDataFile( filename, subdir )
        if xml is None:
            return

        self.demarshaller.demarshall( self.context, xml )

        # XXX: lame. use an adapter.
        self.importFiles( import_context, subdir )

        # XXX: lame. use an adapter.
        self.importReferences( import_context, subdir )

        if not root:
            subdir = '%s/%s' % (subdir, self.context.getId())
        else:
            info( "XMLContentFSImporter: root! subdir=%s" % subdir )

        if not self.folderish:
            return

        existing = self.context.objectIds()

        for object_id, type_name in self.objects( import_context, subdir ):
            info( "XMLContentFSImporter.import: object_id=%s, type_name=%s" % ( object_id, type_name ) )
            if object_id not in existing:
                id = _createObjectByType(type_name, self.context, object_id)
                if id is None:
                    error( "XMLContentFSImporter: Could not create instance of type %s, id %s." % ( type_name, object_id ) )
                    continue
                info( "XMLContentFSImporter: Created object type %s, id %s." % ( type_name, object_id ) )

            object = self.context._getOb(object_id)

            adapter = IGSXMLFilesystemImporter(object)

            # XXX: lame. use Annotations to add the collector. Use an adapter.
            adapter.collector = self.collector
            adapter.import_(import_context, subdir)

    def importFiles( self, import_context, subdir ):
        for field in self.context.Schema().fields():
            if isinstance( field, FileField ):
                mime = field.getContentType( self.context )
                if mime.startswith('text'):
                    continue

                value = import_context.readDataFile( "%s.%s.data"%( self.context.getId(), field.getName()  ), subdir)
                field.set( self.context, value )
                    

    def importReferences( self, import_context, subdir ):
        if self.collector is not None:
            refmap = self.collector.getRefMap()
        else:
            refmap = {}

        # read the binary pickle file
        refs = import_context.readDataFile( "%s.refs.pickle" % self.context.getId(), subdir )
        if refs is None:
            return

        # ... boom! restore references and
        self.context.at_references = pickle.loads( refs )

        # ... fix them up. We make sure that at least the source
        # UID is correct at this time.
        for ref in self.context.at_references.objectValues():
            url = getRelURL( self.context.at_references, ref.getPhysicalPath() )

            # record this change
            refmap[ref.sourceUID] = { "new": self.context.UID(), "old": ref.sourceUID, "target": ref.targetUID, "ref": ref.UID, "url": url }

            # update the reference source
            ref.sourceUID = self.context.UID()
            self.ref_catalog.catalog_object( ref, url )

        return refmap


# vim: set ft=python ts=4 sw=4 expandtab :
